local M = {}

local im = ui_imgui

local tornados = {}
local attachedTornados = {}

local presets = {
    {
        name = "EF 0",
        height = 80,
        radius = 2,
        windRotation = 40,
        windInwards = 37,
        windUp = 37,
        windRange = 30
    },
    {
        name = "EF 1",
        height = 100,
        radius = 2.5,
        windRotation = 50,
        windInwards = 45,
        windUp = 45,
        windRange = 35
    },
    {
        name = "EF 2",
        height = 125,
        radius = 4,
        windRotation = 50,
        windInwards = 60,
        windUp = 60,
        windRange = 45
    },
    {
        name = "EF 3",
        height = 150,
        radius = 6,
        windRotation = 50,
        windInwards = 73,
        windUp = 73,
        windRange = 55
    },
    {
        name = "EF 4",
        height = 200,
        radius = 15,
        windRotation = 60,
        windInwards = 90,
        windUp = 90,
        windRange = 80
    },
    {
        name = "EF 5",
        height = 325,
        radius = 40,
        windRotation = 110,
        windInwards = 150,
        windUp = 150,
        windRange = 100
    },
    {
        name = "EF 6",
        height = 500,
        radius = 75,
        windRotation = 150,
        windInwards = 200,
        windUp = 200,
        windRange = 125
    },
}

local windowOpen = im.BoolPtr(false)
local ui = {
    name = im.ArrayChar(32, "tornado"),
    height = im.FloatPtr(200),
    radius = im.FloatPtr(15),
    windRotation = im.FloatPtr(70),
    windInwards = im.FloatPtr(100),
    windUp = im.FloatPtr(100),
    windRange = im.FloatPtr(70)
}

-- Spawn a tornado
local function spawnTornado(name, pos, height, radiusBottom, radiusTop, windRotation, windInwards, windUp, windRange, isRemote)
    pos = pos or core_camera.getPosition() or vec3(0,0,0)
    height = height or 200
    radiusBottom = radiusBottom or 15
    radiusTop = radiusTop or radiusBottom*8
    windRotation = windRotation or 70
    windInwards = windInwards or 100
    windUp = windUp or 100
    windRange = windRange or 70

    tornados[name] = {
        pos = pos,
        height = height,
        radiusBottom = radiusBottom,
        radiusTop = radiusTop,
        windRotation = windRotation,
        windInwards = windInwards,
        windUp = windUp,
        windRange = windRange
    }

    local objName = name.."_tornado"
    local object = scenetree[objName]
    if not object then
        object = createObject("TSStatic")
        object:setField("shapeName", 0, "/art/Tornado/Tornado.dae")
        object:registerObject(objName)
    end
    object:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)
    object:setScale(vec3(radiusBottom/7, radiusBottom/7, height/110))

    local emitterName = name.."_tornado_particles"
    local emitter = scenetree[emitterName]
    if not emitter then
        emitter = createObject("ParticleEmitterNode")
        emitter:setField("emitter", 0, "TornadoParticle")
        emitter:setField("dataBlock", 0, "lightExampleEmitterNodeData1")
        emitter:registerObject(emitterName)
    end
    emitter:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)

    local sfxName = name.."_tornado_sfx"
    local sfx = scenetree[sfxName]
    if not sfx then
        sfx = createObject("SFXEmitter")
        sfx:setField("fileName", 0, "/art/Tornado/Tornado_Wind.ogg")
        sfx:registerObject(sfxName)
    end
    sfx:preApply()
    sfx:setField("maxDistance", 0, windRange*3)
    sfx:postApply()
    sfx:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)

    be:sendToMailbox("tornados", lpack.encode(tornados))

    -- Trigger server event if it was run locally
    if not isRemote and MPGameNetwork and MPCoreNetwork and MPCoreNetwork.isMPSession() then
        TriggerServerEvent("tornado_spawn", jsonEncode({name, pos, height, radiusBottom, radiusTop, windRotation, windInwards, windUp, windRange}))
    end
end

-- Set data of a tornado
local function setTornadoData(name, data)
    tableMerge(tornados[name], data)
    local tornado = tornados[name]
    local pos = tornado.pos
    
    local object = scenetree[name.."_tornado"]
    if object then
        object:setScale(vec3(tornado.radiusBottom/7, tornado.radiusBottom/7, tornado.height/110))
        object:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)
    end

    local emitter = scenetree[name.."_tornado_particles"]
    if emitter then
        emitter:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)
    end

    local sfx = scenetree[name.."_tornado_sfx"]
    if sfx then
        sfx:preApply()
        sfx:setField("maxDistance", 0, tornado.windRange*3)
        sfx:postApply()
        sfx:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)
    end

    be:sendToMailbox("tornados", lpack.encode(tornados))
end

-- Set only the position of a tornado
local function setTornadoPos(name, pos)
    setTornadoData(name, {pos = pos})
end

-- Attach a tornado to a vehicle
local function setTornadoAttached(name, id, isRemote)
    if attachedTornados[name] then
        local veh = scenetree.findObjectById(attachedTornados[name])
        if veh then
            veh:queueLuaCommand("tornado.setIgnoreTornados(false); obj:setGhostEnabled(false)")
            veh:setMeshAlpha(1, "")
        end
    end
    if id then
        local veh = scenetree.findObjectById(id)
        if veh then
            veh:queueLuaCommand("tornado.setIgnoreTornados(true); obj:setGhostEnabled(true)")
        end
    end
    
    attachedTornados[name] = id

    -- Trigger server event if it was run locally
    if not isRemote and MPGameNetwork and MPCoreNetwork and MPCoreNetwork.isMPSession() then
        TriggerServerEvent("tornado_attach", jsonEncode({name, id and MPVehicleGE.getServerVehicleID(id)}))
    end
end

-- Delete a tornado
local function deleteTornado(name, isRemote)
    local object = scenetree[name.."_tornado"]
    if object then
        object:delete()
    end
    
    local emitter = scenetree[name.."_tornado_particles"]
    if emitter then
        emitter:delete()
    end

    local sfx = scenetree[name.."_tornado_sfx"]
    if sfx then
        sfx:delete()
    end

    setTornadoAttached(name, nil, isRemote)
    tornados[name] = nil

    be:sendToMailbox("tornados", lpack.encode(tornados))

    -- Trigger server event if it was run locally
    if not isRemote and MPGameNetwork and MPCoreNetwork and MPCoreNetwork.isMPSession() then
        TriggerServerEvent("tornado_delete", name)
    end
end

local function remoteSpawnTornado(data)
    local decodedData = jsonDecode(data)

    if decodedData then
        spawnTornado(decodedData[1], vec3(decodedData[2]), decodedData[3], decodedData[4], decodedData[5], decodedData[6], decodedData[7], decodedData[8], decodedData[9], true)
    end
end

local function remoteDeleteTornado(name)
    deleteTornado(name, true)
end

local function remoteAttachTornado(data)
    local decodedData = jsonDecode(data)

    if decodedData then
        local name = decodedData[1]
        local id = decodedData[2]

        setTornadoAttached(name, id and MPVehicleGE.getGameVehicleID(id), true)
    end
end

local function onExtensionLoaded()
    -- Manually load materials from directory since they are not loaded from mod folders
    loadDirRec("art/Tornado")

    if MPGameNetwork then
        AddEventHandler("tornado_spawn", remoteSpawnTornado)
        AddEventHandler("tornado_delete", remoteDeleteTornado)
        AddEventHandler("tornado_attach", remoteAttachTornado)
    end
end

local function onExtensionUnloaded()
    -- Delete all tornados on unload
    for name, tornado in pairs(tornados) do
        deleteTornado(name)
    end
end

local function onBeforeRadialOpened()
    core_quickAccess.addEntry({level = '/root/sandbox/', uniqueID = "tornado", icon = "forceFieldPush1", ignoreAsRecentAction = true, ["goto"] = '/root/sandbox/tornado/', title = "Tornado"})
    
    core_quickAccess.addEntry({
        level = '/root/sandbox/tornado',
        uniqueID = "attachTornado",
        generator = function(entries)
            if not core_input_actionFilter.isActionBlocked("forceField") then
                local e = { title = 'Open GUI', icon = 'beamNG', onSelect = function()
                    tornado.openGui()
                    return { "hide" }
                end }
                table.insert(entries, e)
            end
            if not core_input_actionFilter.isActionBlocked("forceField") then
                for _, preset in ipairs(presets) do
                    local e = { title = 'Attach '..preset.name, icon = 'forceFieldPush1', onSelect = function()
                        local vehID = "tornado_"..tostring(be:getPlayerVehicleID(0))
                        spawnTornado(vehID, nil, preset.height, preset.radius, nil, preset.windRotation, preset.windInwards, preset.windUp, preset.windRange)
                        setTornadoAttached(vehID, be:getPlayerVehicleID(0))
                        return { "hide" }
                    end }
                    --if attachedTornados["tornado_"..tostring(be:getPlayerVehicleID(0))] then e.color = '#ff6600' end
                    table.insert(entries, e)
                end
            end
            if not core_input_actionFilter.isActionBlocked("forceField") then
                local e = { title = 'Delete Tornado', icon = 'beamNG', onSelect = function()
                    local vehID = "tornado_"..tostring(be:getPlayerVehicleID(0))
                    if attachedTornados[vehID] then
                        deleteTornado(vehID)
                    end
                    return { "hide" }
                end }
                if attachedTornados["tornado_"..tostring(be:getPlayerVehicleID(0))] then e.color = '#ff6600' end
                table.insert(entries, e)
            end
        end
    })
end

local function onPreRender(dt)
    if windowOpen[0] and im.Begin("Tornado Settings", windowOpen) then
        if im.BeginMenu("Active Tornados") then
            for name, tornado in pairs(tornados) do
                if im.Selectable1(name) then
                    ui.name = im.ArrayChar(32, name)
                    ui.height[0] = tornado.height
                    ui.radius[0] = tornado.radiusBottom
                    ui.windRotation[0] = tornado.windRotation
                    ui.windInwards[0] = tornado.windInwards
                    ui.windUp[0] = tornado.windUp
                    ui.windRange[0] = tornado.windRange
                end
            end
            im.EndMenu()
        end
        im.Separator()
        
        im.InputText("Name", ui.name)
        
        im.Separator()
        if im.BeginMenu("Presets") then
            for _, preset in ipairs(presets) do
                if im.Selectable1(preset.name) then
                    ui.height[0] = preset.height
                    ui.radius[0] = preset.radius
                    ui.windRotation[0] = preset.windRotation
                    ui.windInwards[0] = preset.windInwards
                    ui.windUp[0] = preset.windUp
                    ui.windRange[0] = preset.windRange
                end
            end
            im.EndMenu()
        end
        im.Separator()

        im.InputFloat("Height", ui.height, 1)
        im.InputFloat("Radius", ui.radius, 1)
        im.InputFloat("Wind Rotation", ui.windRotation, 1)
        im.InputFloat("Wind Inwards", ui.windInwards, 1)
        im.InputFloat("Wind Up", ui.windUp, 1)
        im.InputFloat("Wind Range", ui.windRange, 1)

        im.Separator()
        
        im.Columns(3)
        if im.Button("Spawn Tornado") then
            spawnTornado(ffi.string(ui.name), core_camera.getPosition(), ui.height[0], ui.radius[0], nil, ui.windRotation[0], ui.windInwards[0], ui.windUp[0], ui.windRange[0])
        end
        im.NextColumn()
        if im.Button("Delete Tornado") then
            deleteTornado(ffi.string(ui.name))
        end
        im.NextColumn()
        if im.Button("Attach/Detach Tornado") then
            local name = ffi.string(ui.name)
            if attachedTornados[name] then
                setTornadoAttached(name, nil)
            else
                setTornadoAttached(name, be:getPlayerVehicleID(0))
            end
        end
        im.End()
    end
    
    -- Move attached tornados
    for name, id in pairs(attachedTornados) do
        local veh = scenetree.findObjectById(id)
        if veh then
            setTornadoPos(name, veh:getPosition())
            veh:setMeshAlpha(0.1, "")
        end
    end

    -- Place Sound Emitter close to the camera inside the tornado wall
    local camPos = core_camera.getPosition()
    for name, tornado in pairs(tornados) do
        local sfx = scenetree[name.."_tornado_sfx"]
        if sfx then
            local posDiff = camPos - tornado.pos
            local radius = linearScale(posDiff.z, 0, tornado.height, tornado.radiusBottom, tornado.radiusTop)

            local closestCenterPoint = tornado.pos:copy()
            closestCenterPoint.z = clamp(camPos.z, tornado.pos.z, tornado.pos.z + tornado.height)
            
            local pos = closestCenterPoint + (camPos-closestCenterPoint):normalized()*radius
            sfx:setPosRot(pos.x, pos.y, pos.z, 0, 0, 0, 1)
        end
    end
end

local function onVehicleSpawned(vehId)
    for name, id in pairs(attachedTornados) do
        if id == vehId then
            local veh = scenetree.findObjectById(vehId)
            if veh then
                veh:queueLuaCommand("tornado.setIgnoreTornados(true); obj:setGhostEnabled(true)")
            end
        end
    end
end

local function openGui()
    windowOpen = im.BoolPtr(true)
end

-- public interface
M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded
M.onPreRender = onPreRender
M.spawnTornado = spawnTornado
M.setTornadoData = setTornadoData
M.setTornadoPos = setTornadoPos
M.deleteTornado = deleteTornado
M.setTornadoAttached = setTornadoAttached
M.onBeforeRadialOpened = onBeforeRadialOpened
M.openGui = openGui
M.onVehicleSpawned = onVehicleSpawned

return M