local M = {}

local upVec = vec3(0,0,1)
local ignoreTornados = false

local function setTornados(newTornados)
    tornados = newTornados
end

local function updateGFX(dt)
    local tornados = lpack.decode(obj:getLastMailbox("tornados")) or {}
    
    if ignoreTornados then
        obj:setWind(0, 0, 0)
        return
    end

    local vehPos = obj:getCenterPosition()
    
    local wind = vec3(0,0,0)
    for k, tornado in pairs(tornados) do
        local posDiff = vehPos - tornado.pos
        local radius = linearScale(posDiff.z, 0, tornado.height, tornado.radiusBottom, tornado.radiusTop)
        
        local closestCenterPoint = tornado.pos:copy()
        closestCenterPoint.z = clamp(vehPos.z, tornado.pos.z, tornado.pos.z + tornado.height)

        local distToWall = math.abs(vehPos:distance(closestCenterPoint) - radius)
        local windMul = linearScale(distToWall, 0, tornado.windRange, 1, 0)

        local pullPoint = closestCenterPoint + posDiff:cross(upVec):normalized()*radius
        wind = wind + (pullPoint - vehPos):normalized()*tornado.windRotation*windMul + (closestCenterPoint - vehPos):normalized()*tornado.windInwards*windMul + upVec*tornado.windUp*windMul

        --obj.debugDrawProxy:drawCylinder(tornado.pos, tornado.pos + upVec*0.1, tornado.radiusBottom, color(0, 0, 255, 255))
        --obj.debugDrawProxy:drawCylinder(tornado.pos + upVec*tornado.height, tornado.pos + upVec*tornado.height + upVec*0.1, tornado.radiusTop, color(0, 0, 255, 255))
        --obj.debugDrawProxy:drawSphere(5, closestCenterPoint, color(0, 0, 255, 255))
        --obj.debugDrawProxy:drawSphere(5, pullPoint, color(0, 0, 255, 255))
    end

    obj:setWind(wind.x, wind.y, wind.z)
end

local function onExtensionLoaded()
    --setTornados({
    --    {
    --        pos = vec3(-100,0,100),
    --        height = 100,
    --        radiusBottom = 10,
    --        radiusTop = 50,
    --        windRotation = 100,
    --        windInwards = 150,
    --        windUp = 150,
    --        windRange = 50
    --    }
    --})
end

local function setIgnoreTornados(newIgnoreTornados)
    ignoreTornados = newIgnoreTornados
end

-- public interface
M.onExtensionLoaded = onExtensionLoaded
M.updateGFX = updateGFX
M.setTornados = setTornados
M.setIgnoreTornados = setIgnoreTornados

return M