local M = {}
--Mandatory controller parameters
M.type = "main"
M.defaultOrder = 500

M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0
-----

local atan2 = math.atan2
local pi = math.pi

local pitchPID = newPIDParallel(0.5, 0.5, 0.05, -1, 1)
local rollPID = newPIDParallel(0.5, 0.5, 0.05, -1, 1)
local yawPID = newPIDParallel(0.5, 0.5, 0.05, -1, 1)
local heightPID = newPIDParallel(0.5, 0.5, 0.1, -1, 1)

local walkForce = 20
local sprintForce = 50
local drag = 0.5
local hoverHeight = 0.3

local camPos = vec3()
local camRot = quat()

local moveFBInput = 0
local moveLRInput = 0
local sprintInput = 0
local yawInput = 0
local desiredForce = vec3()

local targetObjectID = -1
local aiMode = "chase"

local upVec = vec3(0,0,1)

local function shortestAngle(target, heading)
    return (target - heading + 3*pi) % (2*pi) - pi
end

local function setCameraRotation(pos, rot)
	camPos = pos
	camRot = rot
end

local function getCameraRotation()
	obj:queueGameEngineLua('be:queueObjectLua('..obj:getID()..', "controller.getControllerSafe(\''..M.name..'\').setCameraRotation("..serialize(core_camera.getPosition())..", "..serialize(core_camera.getQuat())..")")')
end

local function init(jbeamData)
    walkForce = jbeamData.walkForce or walkForce
    sprintForce = jbeamData.sprintForce or sprintForce
    drag = jbeamData.drag or drag
    hoverHeight = jbeamData.hoverHeight or hoverHeight

    input.state["ft_moveF"] =  {val = 0, filter = 0, smootherKBD = newTemporalSmoothing(3, 3, 1000, 0), smootherPAD = newTemporalSmoothing(100, 100, nil, 0), minLimit =  0, maxLimit = 1}
	input.state["ft_moveB"] =  {val = 0, filter = 0, smootherKBD = newTemporalSmoothing(3, 3, 1000, 0), smootherPAD = newTemporalSmoothing(100, 100, nil, 0), minLimit =  0, maxLimit = 1}
	input.state["ft_moveL"] =  {val = 0, filter = 0, smootherKBD = newTemporalSmoothing(3, 3, 1000, 0), smootherPAD = newTemporalSmoothing(100, 100, nil, 0), minLimit =  0, maxLimit = 1}
	input.state["ft_moveR"] =  {val = 0, filter = 0, smootherKBD = newTemporalSmoothing(3, 3, 1000, 0), smootherPAD = newTemporalSmoothing(100, 100, nil, 0), minLimit =  0, maxLimit = 1}
	input.state["ft_sprint"] = {val = 0, filter = 0, smootherKBD = newTemporalSmoothing(3, 3, 1000, 0), smootherPAD = newTemporalSmoothing(100, 100, nil, 0), minLimit =  0, maxLimit = 1}
end

local function reset()

end

local function updateFixedStep(dt)
    local roll, pitch, yaw = obj:getRollPitchYaw()
    local vehPos = obj:getPosition()
    local vehVel = obj:getVelocity()
    local dirRight = obj:getDirectionVectorRight()
    local dirFwd = obj:getDirectionVector()
    local dirUp = obj:getDirectionVectorUp()

    local height = obj:castRayStatic(vehPos+upVec*0.1, -upVec, 0.1+hoverHeight*2)-0.1

    local force = desiredForce - vehVel*drag + heightPID:get(height-hoverHeight, 0, dt)*upVec*100
    local torque = dirRight*pitchPID:get(-pitch, 0, dt) + dirFwd*rollPID:get(roll, 0, dt) + dirUp*yawPID:get(yawInput, 0, dt)
    obj:applyClusterLinearAngularAccel(v.data.refNodes and v.data.refNodes[0].ref or 0, force*dt*2000, -torque*dt*2000000)
end

local function updateGFX(dt)
    -- Inputs (separated to make keyboard inputs work properly)
	moveFBInput = (electrics.values["ft_moveF_input"] or 0) - (electrics.values["ft_moveB_input"] or 0)
	moveLRInput = (electrics.values["ft_moveR_input"] or 0) - (electrics.values["ft_moveL_input"] or 0)
    sprintInput = (electrics.values["ft_sprint_input"] or 0)

    local roll, pitch, yaw = obj:getRollPitchYaw()
    yawInput = 0---yaw
    
    if ai.isDriving() then
        moveFBInput = clamp(input.throttle - input.brake, -1, 1)
		yawInput = clamp(-input.steering, -1, 1)

        desiredForce = obj:getDirectionVector()*moveFBInput*walkForce
    elseif mapmgr.getObjects()[targetObjectID] then
        local targetVec = mapmgr.objects[targetObjectID].pos - obj:getPosition()
        local targetYaw = atan2(targetVec.x, targetVec.y)

        yawInput = clamp(shortestAngle(targetYaw, yaw), -1, 1)
        desiredForce = targetVec:normalized()*walkForce

        if aiMode == "chase" then
            yawInput = clamp(shortestAngle(-targetYaw + pi, yaw), -1, 1)
            desiredForce = targetVec:normalized()*walkForce
        elseif aiMode == "follow" then
            yawInput = clamp(shortestAngle(-targetYaw + pi, yaw), -1, 1)
            desiredForce = targetVec:normalized() * clamp(targetVec:length()-7, -walkForce, walkForce)
        elseif aiMode == "flee" then
            yawInput = clamp(shortestAngle(-targetYaw, yaw), -1, 1)
            desiredForce = targetVec:normalized()*-walkForce
        end
    elseif playerInfo.anyPlayerSeated then
        getCameraRotation()

        local camYaw = -camRot:toEulerYXZ().x + pi
        yawInput = clamp(shortestAngle(camYaw, yaw), -1, 1)
        
        desiredForce = (obj:getDirectionVector()*moveFBInput + obj:getDirectionVectorRight()*moveLRInput)*linearScale(sprintInput, 0, 1, walkForce, sprintForce)
    end
end

local function update(dt)

end

local function target(id, mode)
    targetObjectID = id
    aiMode = mode or "chase"
end

M.init = init
M.reset = reset
M.updateFixedStep = updateFixedStep
M.updateGFX = updateGFX
M.update = update

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setGearboxMode = nop
M.smartParkingBrake = nop
M.setStarter = nop
M.setEngineIgnition = nop
M.setFreeze = nop
M.sendTorqueData = nop
M.vehicleActivated = nop
-------------------------------

M.setCameraRotation = setCameraRotation
M.target = target

return M